"""Provides a message to signify a location change."""

##############################################################################
# Python imports.
from dataclasses import dataclass

##############################################################################
# Textual imports.
from textual.message import Message
from textual.widget import Widget

##############################################################################
# Local imports.
from ..types import Location


##############################################################################
@dataclass
class LocationChanged(Message):
    """Message that can be sent to signify a location change."""

    changer: Widget
    """The widget responsible for changing the location."""
    location: Location
    """The new location."""

    @property
    def control(self) -> Widget:
        """An alias for `changer`."""
        return self.changer


### location.py ends here



"""Widget for showing the AST of some Python code."""

##############################################################################
# Python imports.
from ast import AST, AsyncFunctionDef, ClassDef, FunctionDef, parse
from functools import singledispatchmethod
from typing import Any, Self
from webbrowser import open_new

##############################################################################
# Rich imports.
from rich.markup import escape

##############################################################################
# Textual imports.
from textual import on
from textual.reactive import var
from textual.widgets import Tree
from textual.widgets.tree import TreeNode

##############################################################################
# Textual enhanced imports.
from textual_enhanced.binding import HelpfulBinding

##############################################################################
# Local imports.
from ..messages import LocationChanged
from ..types import Location

##############################################################################
ASTNode = TreeNode[Any]
"""The type of a node in the widget."""


##############################################################################
class AbstractSyntaxTree(Tree[Any]):
    """Widget that displays Python code AST."""

    ICON_NODE = "…"
    ICON_NODE_EXPANDED = " "

    DEFAULT_CSS = """
    AbstractSyntaxTree.--error {
        color: $text-error;
        background: $error 25%;
    }
    """

    BINDINGS = [
        HelpfulBinding(
            "a",
            "about",
            "About AST",
            tooltip="Show the AST entry's documentation in the Python documentation",
        )
    ]

    HELP = """
    ## Abstract Syntax Tree

    This panel is the abstract syntax tree of the Python source code.

    The following keys can be used as shortcuts in this panel:
    """

    code: var[str | None] = var(None)
    """The code to show the AST of."""

    error: var[bool] = var(False)
    """Is there an error with the code we've been given?"""

    def __init__(
        self, id: str | None = None, classes: str | None = None, disabled: bool = False
    ) -> None:
        """Initialise the object.

        Args:
            id: The ID of the AST widget in the DOM.
            classes: The CSS classes of the AST widget.
            disabled: Whether the AST widget is disabled or not.
        """
        super().__init__("", id=id, classes=classes, disabled=disabled)
        self.show_root = False
        self.border_title = "AST"
        self.show_guides = False
        self.guide_depth = 1
        self._line_to_node_map: dict[int, ASTNode] = {}
        """A map to help translate a line number to a node in the tree."""

    def clear(self) -> Self:
        """Clear down the AST tree."""
        self._line_to_node_map = {}
        return super().clear()

    @classmethod
    def maybe_add(cls, value: Any) -> bool:
        """Does the value look like it should be added to the display?

        Args:
            value: The value to consider.

        Returns:
            `True` if the value should be added, `False` if not.
        """
        return bool(value) if isinstance(value, (list, tuple)) else True

    @classmethod
    def _location_of(cls, node: ASTNode) -> Location | None:
        """Get the location of a node in the AST.

        Args:
            node: The node in the tree to get the location of.

        Returns:
            The location, or `None` if nothing could be worked out.
        """
        if all(
            hasattr(node.data, location_property)
            for location_property in (
                "lineno",
                "col_offset",
                "end_lineno",
                "end_col_offset",
            )
        ):
            return Location(
                getattr(node.data, "lineno", None),
                getattr(node.data, "col_offset", None),
                getattr(node.data, "end_lineno", None),
                getattr(node.data, "end_col_offset", None),
            )
        elif node.parent is not None:
            return cls._location_of(node.parent)
        return None

    @singledispatchmethod
    def _base_node(self, item: Any, to_node: ASTNode) -> ASTNode:
        """Attach a base node.

        Args:
            item: The item to associate with the node.
            to_node: The node to attach to.

        Returns:
            The new node.
        """
        return to_node.add(escape(item.__class__.__name__), data=item)

    @_base_node.register
    def _(self, item: AST, to_node: ASTNode) -> ASTNode:
        """Attach a base node.

        Args:
            item: The item to associate with the node.
            to_node: The node to attach to.

        Returns:
            The new node.
        """
        label = f"{escape(item.__class__.__name__)}"
        if isinstance(item, (ClassDef, FunctionDef, AsyncFunctionDef)):
            label = f"{label} [dim italic]{escape(item.name)}[/]"
        node = to_node.add(label, data=item)
        if (
            location := self._location_of(to_node)
        ) is not None and location.line_number not in self._line_to_node_map:
            self._line_to_node_map[location.line_number] = node
        return node

    @_base_node.register
    def _(self, item: str, to_node: ASTNode) -> ASTNode:
        """Attach a base node.

        Args:
            item: The item to associate with the node.
            to_node: The node to attach to.

        Returns:
            The new node.
        """
        return to_node.add(escape(item), data=item)

    @singledispatchmethod
    def _add(self, item: Any, to_node: ASTNode) -> Self:
        """Add an AST item to the tree.

        Args:
            item: The AST item to add.
            to_node: The node to add it to.
        """
        if isinstance(item, (list, tuple)):
            for child_item in item:
                self._add(child_item, to_node)
        else:
            to_node.add_leaf(escape(repr(item)), data=item)
        return self

    @_add.register
    def _(self, item: AST, to_node: ASTNode) -> Self:
        """Add an AST item to the tree.

        Args:
            item: The ast entry to add.
            to_node: The node to add to.
        """
        node = self._base_node(item, to_node)
        if item._fields:
            for field in item._fields:
                if self.maybe_add(value := getattr(item, field)):
                    self._add(value, self._base_node(field, node))
        else:
            node.allow_expand = False
        return self

    def _watch_error(self) -> None:
        """React to the error state being toggled."""
        self.set_class(self.error, "--error")

    def _watch_code(self) -> None:
        """React to the code being changed."""
        if not self.code:
            self.clear()
            return
        try:
            ast = parse(self.code)
        except SyntaxError:
            self.error = True
            return
        self.error = False
        self.clear()._add(ast, self.root).root.expand_all()
        self.move_cursor(self.root)

    @on(Tree.NodeHighlighted)
    def _ast_node_highlighted(self, message: Tree.NodeHighlighted[ASTNode]) -> None:
        """Handle a node being highlighted in the AST.

        Args:
            message: The message to handle.
        """
        if location := self._location_of(message.node):
            message.stop()
            self.post_message(LocationChanged(self, location))

    def goto_first_node_on_line(self, line: int) -> None:
        """Go to the first node that's related to the given line.

        Args:
            line: The line to find a node for.
        """
        if (node := self._line_to_node_map.get(line)) is not None:
            with self.prevent(Tree.NodeHighlighted):
                self.move_cursor(node)

    def _closest_ast(self, node: ASTNode) -> AST | None:
        """Get the closest AST item to the given tree node.

        Args:
            node: The node to get the AST from.

        Returns:
            The closest AST entry, or `None` if none found.
        """
        if isinstance(node.data, AST):
            return node.data
        return self._closest_ast(node.parent) if node.parent else None

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        """Check if an action is possible to perform right now.

        Args:
            action: The action to perform.
            parameters: The parameters of the action.

        Returns:
            `True` if it can perform, `False` or `None` if not.
        """
        if action == "about":
            return bool(self.root.children)
        return True

    def action_about(self) -> None:
        """Handle a request to view the AST item's documentation."""
        if self.cursor_node and (ast := self._closest_ast(self.cursor_node)):
            open_new(
                f"https://docs.python.org/3/library/ast.html#ast.{ast.__class__.__name__}"
            )


### ast.py ends here



"""Code theme commands for the command palette."""

##############################################################################
from textual.widgets import TextArea

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import CommandHit, CommandHits, CommandsProvider

##############################################################################
# Local imports.
from ..messages import SetCodeTheme


##############################################################################
class CodeThemeCommands(CommandsProvider):
    """A command provider for code themes."""

    def commands(self) -> CommandHits:
        """Provide code editor theme commands for the command palette.

        Yields:
            The commands for the command palette.
        """
        for theme in sorted(TextArea().available_themes):
            yield CommandHit(
                theme,
                f"Change the theme of the code editor to {theme}",
                SetCodeTheme(theme),
            )


### code_themes.py ends here



"""Provides helper types used in the application."""

##############################################################################
# Python imports.
from typing import NamedTuple


##############################################################################
class Location(NamedTuple):
    """A location within a body of code."""

    start_line: int | None
    """The line the location starts on."""
    start_column: int | None = None
    """The column the location starts on."""
    end_line: int | None = None
    """The line the location ends on."""
    end_column: int | None = None
    """The column the location ends on."""

    @property
    def line_number(self) -> int:
        """Alias for `start_line`."""
        return self.start_line or 0

    @property
    def line_number_only(self) -> bool:
        """Do we only have a line number?"""
        return self.start_line is not None and all(
            location is None
            for location in (
                self.start_column,
                self.end_line,
                self.end_column,
            )
        )


### types.py ends here



"""Provides the main application commands for the command palette."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import (
    ChangeTheme,
    CommandHits,
    CommandsProvider,
    Help,
    Quit,
)

##############################################################################
# Local imports.
from ..commands import (
    ChangeCodeTheme,
    LoadFile,
    NewCode,
    ShowASTOnly,
    ShowDisassemblyAndAST,
    ShowDisassemblyOnly,
    SwitchLayout,
    ToggleOffsets,
    ToggleOpcodes,
)


##############################################################################
class MainCommands(CommandsProvider):
    """Provides some top-level commands for the application."""

    def commands(self) -> CommandHits:
        """Provide the main application commands for the command palette.

        Yields:
            The commands for the command palette.
        """
        yield ChangeCodeTheme()
        yield ChangeTheme()
        yield Help()
        yield Quit()
        yield LoadFile()
        yield NewCode()
        yield SwitchLayout()
        yield ToggleOffsets()
        yield ToggleOpcodes()
        yield ShowASTOnly()
        yield ShowDisassemblyAndAST()
        yield ShowDisassemblyOnly()


### main.py ends here



"""Widget for showing some Python source code."""

##############################################################################
# Textual imports.
from textual import on
from textual.widgets import TextArea
from textual.widgets.text_area import Selection

##############################################################################
# Local imports.
from ..messages import LocationChanged
from ..types import Location


##############################################################################
class Source(TextArea):
    """Widget that displays Python source code."""

    HELP = """
    ## Python Source

    This panel is the Python source code that you're exploring.
    """

    def __init__(self) -> None:
        """Initialise the widget."""
        super().__init__(
            "",
            language="python",
            soft_wrap=False,
            show_line_numbers=True,
        )
        self.border_title = "Source"

    def highlight_location(self, location: Location) -> None:
        """Highlight the given location.

        Args:
            location: The location message to get the data from.
        """
        with self.prevent(TextArea.SelectionChanged):
            if location.line_number_only:
                self.select_line(location.line_number - 1)
            elif (
                location.start_line is not None
                and location.start_column is not None
                and location.end_line is not None
                and location.end_column is not None
            ):
                self.selection = Selection(
                    start=(location.start_line - 1, location.start_column),
                    end=(location.end_line - 1, location.end_column),
                )
            else:
                self.selection = Selection.cursor(self.selection.end)
        self.scroll_cursor_visible(True)

    @on(TextArea.SelectionChanged)
    def _cursor_location_changed(self, message: TextArea.SelectionChanged) -> None:
        """Handle the cursor location changing."""
        message.stop()
        self.post_message(LocationChanged(self, Location(message.selection.end[0] + 1)))


### source.py ends here



"""Widget for showing the disassembly of some Python code."""

##############################################################################
# Python imports.
from dis import Bytecode, Instruction, opname
from types import CodeType
from typing import Final, Self
from webbrowser import open_new

##############################################################################
# Rich imports.
from rich.console import Group
from rich.markup import escape
from rich.rule import Rule
from rich.table import Table

##############################################################################
# Textual imports.
from textual import on
from textual.reactive import var
from textual.widgets.option_list import Option, OptionDoesNotExist

##############################################################################
# Textual enhanced imports.
from textual_enhanced.binding import HelpfulBinding
from textual_enhanced.widgets import EnhancedOptionList

##############################################################################
# Local imports.
from ..messages import LocationChanged
from ..types import Location

##############################################################################
LINE_NUMBER_WIDTH: Final[int] = 6
"""Width for line numbers."""
OFFSET_WIDTH: Final[int] = 4
"""The width of the display of the offset."""
OPNAME_WIDTH: Final[int] = max(len(operation) for operation in opname)
"""Get the maximum length of an operation name."""


##############################################################################
class Code(Option):
    """Option that marks a new disassembly."""

    def __init__(self, code: CodeType) -> None:
        """Initialise the object.

        Args:
            code: The code that will follow.
        """
        super().__init__(
            Group("", Rule(f"[dim bold]@{hex(id(code))}[/]", style="dim bold")),
            id=f"{hex(id(code))}",
        )


##############################################################################
class Operation(Option):
    """The view of an operation."""

    def __init__(
        self,
        operation: Instruction,
        *,
        show_offset: bool = False,
        show_opcode: bool = False,
        code: CodeType | None = None,
    ) -> None:
        """Initialise the object.

        Args:
            operation: The operation.
            show_offset: Show the offset in the display?
            show_opcode: Show the opcode in the display?
            code: The code that the operation came from.
        """
        self._operation = operation
        """The operation being displayed."""
        self._code = code
        """The code the operation came from."""

        display = Table.grid(expand=True, padding=1)
        display.add_column(width=LINE_NUMBER_WIDTH)
        display.add_column(width=OFFSET_WIDTH if show_offset else 0)
        display.add_column(width=OPNAME_WIDTH)
        display.add_column(ratio=1)
        display.add_row(
            str(operation.line_number)
            if operation.line_number is not None and operation.starts_line
            else "",
            f"[dim]{operation.offset}[/]",
            f"{operation.opname} [dim]({operation.opcode})[/]"
            if show_opcode
            else operation.opname,
            f"[dim]code@[/]{hex(id(operation.argval))}"
            if isinstance(operation.argval, CodeType)
            # With Python 3.14, LOAD_SMALL_INT at least has no populated
            # argrepr value despite having a value, so here we use argrepr
            # if it has a value, and if not we repr any non-None argval.
            # This could result in a false negative I guess, but I'd hope
            # that mostly argrepr does the right thing.
            else escape(
                operation.argrepr
                or (repr(operation.argval) if operation.argval is not None else "")
            ),
        )
        super().__init__(
            Group(
                Rule(
                    f"[italic dim]-- L{operation.label}[/]",
                    align="left",
                    style="dim",
                    characters="-",
                ),
                display,
            )
            if operation.is_jump_target
            else display,
            id=self.make_id(operation.offset, code),
        )

    @property
    def operation(self) -> Instruction:
        """The operation being displayed."""
        return self._operation

    @property
    def code(self) -> CodeType | None:
        """The code that the operation belongs to."""
        return self._code

    @staticmethod
    def make_id(offset: int, code: CodeType | None = None) -> str | None:
        """Make an ID for the given operation.

        Args:
           offset: The offset of the instruction.
           code: The code the instruction came from.

        Returns:
            The ID for the operation, or [`None`] if one isn't needed.
        """
        if code:
            return f"operation-{hex(id(code))}-{offset}"
        return f"operation-{offset}"


##############################################################################
class Disassembly(EnhancedOptionList):
    """Widget that displays Python code disassembly."""

    DEFAULT_CSS = """
    Disassembly.--error {
        color: $text-error;
        background: $error 25%;
    }
    """

    BINDINGS = [
        HelpfulBinding(
            "a",
            "about",
            "About opcode",
            tooltip="Show the opcode's documentation in the Python documentation",
        )
    ]

    HELP = """
    ## Disassembly

    This panel is the disassembly of the Python source code.

    The following keys can be used as shortcuts in this panel:
    """

    code: var[str | None] = var(None)
    """The code to disassemble."""

    show_offset: var[bool] = var(False, init=False)
    """Show the offset of each instruction?"""

    show_opcodes: var[bool] = var(False, init=False)
    """Should we show the opcodes in the disassembly?"""

    error: var[bool] = var(False)
    """Is there an error with the code we've been given?"""

    def __init__(
        self, id: str | None = None, classes: str | None = None, disabled: bool = False
    ):
        """Initialise the object.

        Args:
            name: The name of the disassembly.
            id: The ID of the disassembly in the DOM.
            classes: The CSS classes of the disassembly.
            disabled: Whether the disassembly is disabled or not.
        """
        super().__init__(id=id, classes=classes, disabled=disabled)
        self._line_map: dict[int, int] = {}
        """A map of line numbers to locations within the disassembly display."""
        self.border_title = "Disassembly"

    def _add_operations(self, code: str | CodeType, fresh: bool = False) -> Self:
        """Add the operations from the given code.

        Args:
            code: The code to add the operations from.
            fresh: Is this a fresh add; should we clear the display?

        Returns:
            Self.
        """

        # Build the code up first.
        try:
            operations = Bytecode(code)
        except SyntaxError:
            # There was an error so nope out, but keep the display as is so
            # the user can see what was and also doesn't keep getting code
            # disappear and then appear again.
            self.error = True
            return self
        self.error = False

        # If this is a fresh add we start out by clearing everything down.
        if fresh:
            self.clear_options()
            self._line_map = {}

        # If we've been given a code object, rather than some source, add a
        # marker for that.
        if isinstance(code, CodeType):
            self.add_option(Code(code))

        # Add each operation...
        for operation in operations:
            self.add_option(
                Operation(
                    operation,
                    show_offset=self.show_offset,
                    show_opcode=self.show_opcodes,
                    code=operations.codeobj,
                )
            )
            if operation.line_number is not None and operation.starts_line:
                self._line_map[operation.line_number] = self.option_count - 1

        # Now look for any operations that have code as their arguments, and
        # add that code too.
        for operation in operations:
            if isinstance(operation.argval, CodeType):
                self._add_operations(operation.argval)

        return self

    def _watch_error(self) -> None:
        """React to the error state being toggled."""
        self.set_class(self.error, "--error")

    def _repopulate(self) -> None:
        """Fully repopulate the display."""
        with self.preserved_highlight:
            self._add_operations(self.code or "", True)

    def _watch_code(self) -> None:
        """React to the code being changed."""
        self._repopulate()

    def _watch_show_offset(self) -> None:
        """React to the show offset flag being toggled."""
        self._repopulate()

    def _watch_show_opcodes(self) -> None:
        """React to the show opcodes flag being toggled."""
        self._repopulate()

    @on(EnhancedOptionList.OptionHighlighted)
    def _instruction_highlighted(
        self, message: EnhancedOptionList.OptionHighlighted
    ) -> None:
        """Handle an instruction being highlighted.

        Args:
            message: The message to handle.
        """
        message.stop()
        if isinstance(message.option, Operation):
            self.post_message(
                LocationChanged(
                    self,
                    Location(message.option.operation.line_number)
                    if (position := message.option.operation.positions) is None
                    else Location(
                        position.lineno,
                        position.col_offset,
                        position.end_lineno,
                        position.end_col_offset,
                    ),
                )
            )

    @on(EnhancedOptionList.OptionSelected)
    def _maybe_jump_to_code(self, message: EnhancedOptionList.OptionSelected) -> None:
        """Maybe jump to a selected bit of code.

        Args:
            message: The message to handle.
        """
        message.stop()
        if isinstance(message.option, Operation):
            if isinstance(message.option.operation.argval, CodeType):
                self.highlighted = self.get_option_index(
                    hex(id(message.option.operation.argval))
                )
            elif message.option.operation.jump_target is not None:
                if jump_id := Operation.make_id(
                    message.option.operation.jump_target, message.option.code
                ):
                    try:
                        self.highlighted = self.get_option_index(jump_id)
                    except OptionDoesNotExist:
                        self.notify(
                            "Unable to find that jump location",
                            title="Error",
                            severity="error",
                        )

    def goto_first_instruction_on_line(self, line: int) -> None:
        """Go to the first instruction for a given line number.

        Args:
            line: The line number to find the first instruction for.
        """
        if line in self._line_map:
            with self.prevent(EnhancedOptionList.OptionHighlighted):
                self.highlighted = self._line_map[line]

    def action_about(self) -> None:
        """Handle a request to view the opcode's documentation."""
        if self.highlighted is not None and isinstance(
            option := self.get_option_at_index(self.highlighted), Operation
        ):
            open_new(
                f"https://docs.python.org/3/library/dis.html#opcode-{option.operation.opname}"
            )


### disassembly.py ends here



"""Provides functions and classes for managing the app's data."""

##############################################################################
# Local imports.
from .config import (
    Configuration,
    load_configuration,
    save_configuration,
    update_configuration,
)

##############################################################################
# Exports.
__all__ = [
    "Configuration",
    "load_configuration",
    "save_configuration",
    "update_configuration",
]

### __init__.py ends here



"""Provides the main commands for the application."""

##############################################################################
# Local imports.
from .disassembly import ToggleOffsets, ToggleOpcodes
from .main import (
    ChangeCodeTheme,
    LoadFile,
    NewCode,
    ShowASTOnly,
    ShowDisassemblyAndAST,
    ShowDisassemblyOnly,
    SwitchLayout,
)

##############################################################################
# Exports.
__all__ = [
    "ChangeCodeTheme",
    "LoadFile",
    "NewCode",
    "ShowASTOnly",
    "ShowDisassemblyAndAST",
    "ShowDisassemblyOnly",
    "SwitchLayout",
    "ToggleOffsets",
    "ToggleOpcodes",
]


### __init__.py ends here



"""The main screen."""

##############################################################################
# Python imports.
from argparse import Namespace
from pathlib import Path
from platform import python_version

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.reactive import var
from textual.widgets import Footer, Header

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import ChangeTheme, Command, Help, Quit
from textual_enhanced.screen import EnhancedScreen

##############################################################################
# Textual fspicker imports.
from textual_fspicker import FileOpen, Filters

##############################################################################
# Local imports.
from .. import __version__
from ..commands import (
    ChangeCodeTheme,
    LoadFile,
    NewCode,
    ShowASTOnly,
    ShowDisassemblyAndAST,
    ShowDisassemblyOnly,
    SwitchLayout,
    ToggleOffsets,
    ToggleOpcodes,
)
from ..data import load_configuration, update_configuration
from ..messages import LocationChanged, SetCodeTheme
from ..providers import CodeThemeCommands, MainCommands
from ..widgets import AbstractSyntaxTree, Disassembly, Source


##############################################################################
class Main(EnhancedScreen[None]):
    """The main screen for the application."""

    TITLE = f"DHV v{__version__}"
    SUB_TITLE = f"Python {python_version()}"

    DEFAULT_CSS = """
    Main.--horizontal {
        layout: horizontal;
    }

    Source, Disassembly, AbstractSyntaxTree {
        width: 1fr;
        height: 1fr;
        border: none;
        border-top: solid $panel;
        border-left: solid $panel;
        border-title-color: $border 75%;
        padding-right: 0;
        &:focus {
            border: none;
            border-top: solid $border;
            border-left: solid $border;
            border-title-color: $border;
            background: $panel 80%;
        }
        &.--hidden {
            display: none;
        }
    }
    """

    HELP = """
    ## Main application keys and commands

    The following key bindings and commands are available:
    """

    COMMAND_MESSAGES = (
        # Keep these together as they're bound to function keys and destined
        # for the footer.
        Help,
        Quit,
        NewCode,
        LoadFile,
        # Everything else.
        ChangeCodeTheme,
        ChangeTheme,
        SwitchLayout,
        ToggleOffsets,
        ToggleOpcodes,
        ShowASTOnly,
        ShowDisassemblyAndAST,
        ShowDisassemblyOnly,
    )

    BINDINGS = Command.bindings(*COMMAND_MESSAGES)

    COMMANDS = {MainCommands}

    code: var[str | None] = var(None)
    """The code we're viewing."""

    horizontal_layout: var[bool] = var(True)
    """Should the panes lay out horizontally?"""

    show_disassembly: var[bool] = var(True)
    """Should we show the disassembly panel?"""

    show_ast: var[bool] = var(False)
    """Should we show the AST panel?"""

    def __init__(self, arguments: Namespace) -> None:
        """Initialise the main screen.

        Args:
            arguments: The arguments passed to the application on the command line.
        """
        self._arguments = arguments
        """The arguments passed on the command line."""
        super().__init__()

    def compose(self) -> ComposeResult:
        """Compose the content of the screen."""
        yield Header()
        yield Source()
        with Vertical():
            yield Disassembly().data_bind(Main.code)
            yield AbstractSyntaxTree().data_bind(Main.code)
        yield Footer()

    def _show_source(self, source: Path) -> None:
        """Load up the content of a Python source file.

        Args:
            source: The path to the source file to load.
        """
        try:
            self.query_one(Source).load_text(source.read_text())
        except IOError as error:
            self.notify(str(error), title=f"Unable to load {source}", severity="error")
            return
        with update_configuration() as config:
            config.last_load_location = str(source.absolute().parent)

    def on_mount(self) -> None:
        """Configure the display once the DOM is mounted."""
        config = load_configuration()
        self.horizontal_layout = config.horizontal_layout
        self.show_ast = config.show_ast
        self.show_disassembly = config.show_disassembly
        self.query_one(Disassembly).show_offset = config.show_offsets
        self.query_one(Disassembly).show_opcodes = config.show_opcodes
        self.query_one(Source).theme = config.code_theme or "css"
        if isinstance(to_open := self._arguments.source, Path):
            self._show_source(to_open)

    def _watch_horizontal_layout(self) -> None:
        """React to the horizontal layout setting being changed."""
        self.set_class(self.horizontal_layout, "--horizontal")

    def _watch_show_disassembly(self) -> None:
        """React to the disassembly visibility state change."""
        self.query_one(Disassembly).set_class(not self.show_disassembly, "--hidden")

    def _watch_show_ast(self) -> None:
        """React to the AST visibility state change."""
        self.query_one(AbstractSyntaxTree).set_class(not self.show_ast, "--hidden")

    @on(LocationChanged)
    def _location_changed(self, message: LocationChanged) -> None:
        """React to a change of highlighted instruction in the code.

        Args:
            message: The message to handle.
        """
        # If we're not in the source, or we're not getting an update from
        # the source, update the source.
        if not isinstance(message.changer, Source) and not isinstance(
            self.focused, Source
        ):
            self.query_one(Source).highlight_location(message.location)

        # If we're not in the disassembly, or we're not getting an update
        # from the disassembly, update the disassembly.
        if (
            not isinstance(message.changer, Disassembly)
            and not isinstance(self.focused, Disassembly)
            and message.location.start_line is not None
        ):
            self.query_one(Disassembly).goto_first_instruction_on_line(
                message.location.start_line
            )

        # If we're not in the AST, or we're not getting an update from the
        # AST, update the AST.
        if (
            not isinstance(message.changer, AbstractSyntaxTree)
            and not isinstance(self.focused, AbstractSyntaxTree)
            and message.location.start_line is not None
        ):
            self.query_one(AbstractSyntaxTree).goto_first_node_on_line(
                message.location.start_line
            )

    @on(Source.Changed)
    def _code_changed(self) -> None:
        """Handle the fact that the code has changed."""
        self.code = self.query_one(Source).document.text

    def action_new_code_command(self) -> None:
        """Handle the new code command."""
        self.query_one(Source).load_text("")

    @work
    async def action_load_file_command(self) -> None:
        """Browse for and open a Python source file."""
        if not (
            start_location := Path(load_configuration().last_load_location or ".")
        ).is_dir():
            start_location = Path(".")
        if python_file := await self.app.push_screen_wait(
            FileOpen(
                location=str(start_location),
                title="Load Python code",
                open_button="Load",
                must_exist=True,
                filters=Filters(
                    (
                        "Python",
                        lambda p: p.suffix.lower() in (".py", ".pyi", ".pyw", ".py3"),
                    ),
                    ("All", lambda _: True),
                ),
            )
        ):
            self._show_source(python_file)

    def action_switch_layout_command(self) -> None:
        """Switch the layout of the window."""
        self.horizontal_layout = not self.horizontal_layout
        with update_configuration() as config:
            config.horizontal_layout = self.horizontal_layout

    def action_toggle_offsets_command(self) -> None:
        """Toggle the display of the offsets."""
        show = not self.query_one(Disassembly).show_offset
        self.query_one(Disassembly).show_offset = show
        with update_configuration() as config:
            config.show_offsets = show

    def action_toggle_opcodes_command(self) -> None:
        """Toggle the display of the numeric opcodes."""
        show = not self.query_one(Disassembly).show_opcodes
        self.query_one(Disassembly).show_opcodes = show
        with update_configuration() as config:
            config.show_opcodes = show

    def _save_panels(self) -> None:
        """Remember which panels are visible."""
        with update_configuration() as config:
            config.show_ast = self.show_ast
            config.show_disassembly = self.show_disassembly

    def action_show_disassembly_only_command(self) -> None:
        """Show only the disassembly."""
        self.show_disassembly = True
        self.show_ast = False
        self._save_panels()

    def action_show_ast_only_command(self) -> None:
        """Show only the AST."""
        self.show_disassembly = False
        self.show_ast = True
        self._save_panels()

    def action_show_disassembly_and_ast_command(self) -> None:
        """Show both the disassembly and the AST."""
        self.show_disassembly = True
        self.show_ast = True
        self._save_panels()

    def action_change_code_theme_command(self) -> None:
        """Change the theme used for the code editor."""
        self.show_palette(CodeThemeCommands)

    @on(SetCodeTheme)
    def _set_code_theme(self, message: SetCodeTheme) -> None:
        """Set the theme used by the code editor.

        Args:
            message: The requesting the theme change.
        """
        self.query_one(Source).theme = message.theme
        if message.theme == "css":  # https://github.com/Textualize/textual/issues/5964
            self.query_one(Source).styles.background = None
        with update_configuration() as config:
            config.code_theme = message.theme


### main.py ends here
